@@ -7,7 +7,7 @@
  * And contributors:
  * Nabil Sayegh <postgresql@e-trolley.de>
  *
- * Copyright (c) 2002-2005, PostgreSQL Global Development Group
+ * Copyright (c) 2002-2006, PostgreSQL Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose, without fee, and without a written agreement
@@ -42,6 +42,8 @@
 
 #include "tablefunc.h"
 
+PG_MODULE_MAGIC;
+
 static int	load_categories_hash(char *cats_sql, MemoryContext per_query_ctx);
 static Tuplestorestate *get_crosstab_tuplestore(char *sql,
 						int num_categories,
@@ -854,6 +856,7 @@ get_crosstab_tuplestore(char *sql,
 	HeapTuple	tuple;
 	int			ret;
 	int			proc;
+	MemoryContext SPIcontext;
 
 	/* initialize our tuplestore */
 	tupstore = tuplestore_begin_heap(true, false, work_mem);
@@ -916,9 +919,9 @@ get_crosstab_tuplestore(char *sql,
 			ereport(ERROR,
 					(errcode(ERRCODE_SYNTAX_ERROR),
 					 errmsg("invalid return type"),
-					 errdetail("query-specified return " \
+					 errdetail("Query-specified return " \
 							   "tuple has %d columns but crosstab " \
-							   "returns %d", tupdesc->natts, result_ncols)));
+							   "returns %d.", tupdesc->natts, result_ncols)));
 
 		/* allocate space */
 		values = (char **) palloc(result_ncols * sizeof(char *));
@@ -952,7 +955,12 @@ get_crosstab_tuplestore(char *sql,
 				{
 					/* rowid changed, flush the previous output row */
 					tuple = BuildTupleFromCStrings(attinmeta, values);
+
+					/* switch to appropriate context while storing the tuple */
+					SPIcontext = MemoryContextSwitchTo(per_query_ctx);
 					tuplestore_puttuple(tupstore, tuple);
+					MemoryContextSwitchTo(SPIcontext);
+
 					for (j = 0; j < result_ncols; j++)
 						xpfree(values[j]);
 				}
@@ -983,7 +991,11 @@ get_crosstab_tuplestore(char *sql,
 
 		/* flush the last output row */
 		tuple = BuildTupleFromCStrings(attinmeta, values);
+
+		/* switch to appropriate context while storing the tuple */
+		SPIcontext = MemoryContextSwitchTo(per_query_ctx);
 		tuplestore_puttuple(tupstore, tuple);
+		MemoryContextSwitchTo(SPIcontext);
 	}
 
 	if (SPI_finish() != SPI_OK_FINISH)
@@ -1265,13 +1277,11 @@ build_tuplestore_recursively(char *key_fld,
 							 Tuplestorestate *tupstore)
 {
 	TupleDesc	tupdesc = attinmeta->tupdesc;
-	StringInfo	sql = makeStringInfo();
+	MemoryContext oldcontext;
 	int			ret;
 	int			proc;
 	int			serial_column;
-	StringInfo	branchstr = NULL;
-	StringInfo	chk_branchstr = NULL;
-	StringInfo	chk_current_key = NULL;
+	StringInfoData sql;
 	char	  **values;
 	char	   *current_key;
 	char	   *current_key_parent;
@@ -1283,17 +1293,12 @@ build_tuplestore_recursively(char *key_fld,
 	if (max_depth > 0 && level > max_depth)
 		return tupstore;
 
-	/* start a new branch */
-	branchstr = makeStringInfo();
-
-	/* need these to check for recursion */
-	chk_branchstr = makeStringInfo();
-	chk_current_key = makeStringInfo();
+	initStringInfo(&sql);
 
 	/* Build initial sql statement */
 	if (!show_serial)
 	{
-		appendStringInfo(sql, "SELECT %s, %s FROM %s WHERE %s = %s AND %s IS NOT NULL AND %s <> %s",
+		appendStringInfo(&sql, "SELECT %s, %s FROM %s WHERE %s = %s AND %s IS NOT NULL AND %s <> %s",
 						 key_fld,
 						 parent_key_fld,
 						 relname,
@@ -1304,7 +1309,7 @@ build_tuplestore_recursively(char *key_fld,
 	}
 	else
 	{
-		appendStringInfo(sql, "SELECT %s, %s FROM %s WHERE %s = %s AND %s IS NOT NULL AND %s <> %s ORDER BY %s",
+		appendStringInfo(&sql, "SELECT %s, %s FROM %s WHERE %s = %s AND %s IS NOT NULL AND %s <> %s ORDER BY %s",
 						 key_fld,
 						 parent_key_fld,
 						 relname,
@@ -1350,15 +1355,21 @@ build_tuplestore_recursively(char *key_fld,
 		/* construct the tuple */
 		tuple = BuildTupleFromCStrings(attinmeta, values);
 
+		/* switch to long lived context while storing the tuple */
+		oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
 		/* now store it */
 		tuplestore_puttuple(tupstore, tuple);
 
+		/* now reset the context */
+		MemoryContextSwitchTo(oldcontext);
+
 		/* increment level */
 		level++;
 	}
 
 	/* Retrieve the desired rows */
-	ret = SPI_execute(sql->data, true, 0);
+	ret = SPI_execute(sql.data, true, 0);
 	proc = SPI_processed;
 
 	/* Check for qualifying tuples */
@@ -1368,6 +1379,9 @@ build_tuplestore_recursively(char *key_fld,
 		SPITupleTable *tuptable = SPI_tuptable;
 		TupleDesc	spi_tupdesc = tuptable->tupdesc;
 		int			i;
+		StringInfoData branchstr;
+		StringInfoData chk_branchstr;
+		StringInfoData chk_current_key;
 
 		/* First time through, do a little more setup */
 		if (level == 0)
@@ -1388,28 +1402,35 @@ build_tuplestore_recursively(char *key_fld,
 
 		for (i = 0; i < proc; i++)
 		{
+			/* start a new branch */
+			initStringInfo(&branchstr);
+
+			/* need these to check for recursion */
+			initStringInfo(&chk_branchstr);
+			initStringInfo(&chk_current_key);
+
 			/* initialize branch for this pass */
-			appendStringInfo(branchstr, "%s", branch);
-			appendStringInfo(chk_branchstr, "%s%s%s", branch_delim, branch, branch_delim);
+			appendStringInfo(&branchstr, "%s", branch);
+			appendStringInfo(&chk_branchstr, "%s%s%s", branch_delim, branch, branch_delim);
 
 			/* get the next sql result tuple */
 			spi_tuple = tuptable->vals[i];
 
 			/* get the current key and parent */
 			current_key = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
-			appendStringInfo(chk_current_key, "%s%s%s", branch_delim, current_key, branch_delim);
+			appendStringInfo(&chk_current_key, "%s%s%s", branch_delim, current_key, branch_delim);
 			current_key_parent = pstrdup(SPI_getvalue(spi_tuple, spi_tupdesc, 2));
 
 			/* get the current level */
 			sprintf(current_level, "%d", level);
 
 			/* check to see if this key is also an ancestor */
-			if (strstr(chk_branchstr->data, chk_current_key->data))
+			if (strstr(chk_branchstr.data, chk_current_key.data))
 				elog(ERROR, "infinite recursion detected");
 
 			/* OK, extend the branch */
-			appendStringInfo(branchstr, "%s%s", branch_delim, current_key);
-			current_branch = branchstr->data;
+			appendStringInfo(&branchstr, "%s%s", branch_delim, current_key);
+			current_branch = branchstr.data;
 
 			/* build a tuple */
 			values[0] = pstrdup(current_key);
@@ -1431,9 +1452,15 @@ build_tuplestore_recursively(char *key_fld,
 			xpfree(current_key);
 			xpfree(current_key_parent);
 
+			/* switch to long lived context while storing the tuple */
+			oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
 			/* store the tuple for later use */
 			tuplestore_puttuple(tupstore, tuple);
 
+			/* now reset the context */
+			MemoryContextSwitchTo(oldcontext);
+
 			heap_freetuple(tuple);
 
 			/* recurse using current_key_parent as the new start_with */
@@ -1454,14 +1481,9 @@ build_tuplestore_recursively(char *key_fld,
 													tupstore);
 
 			/* reset branch for next pass */
-			xpfree(branchstr->data);
-			initStringInfo(branchstr);
-
-			xpfree(chk_branchstr->data);
-			initStringInfo(chk_branchstr);
-
-			xpfree(chk_current_key->data);
-			initStringInfo(chk_current_key);
+			xpfree(branchstr.data);
+			xpfree(chk_branchstr.data);
+			xpfree(chk_current_key.data);
 		}
 	}
 
@@ -1524,12 +1546,12 @@ validateConnectbyTupleDesc(TupleDesc tupdesc, bool show_branch, bool show_serial
 
 	/* check that the type of the fifth column is INT4 */
 	if (show_branch && show_serial && tupdesc->attrs[4]->atttypid != INT4OID)
-		elog(ERROR, "Query-specified return tuple not valid for Connectby: "
+		elog(ERROR, "query-specified return tuple not valid for Connectby: "
 			 "fifth column must be type %s", format_type_be(INT4OID));
 
 	/* check that the type of the fifth column is INT4 */
 	if (!show_branch && show_serial && tupdesc->attrs[3]->atttypid != INT4OID)
-		elog(ERROR, "Query-specified return tuple not valid for Connectby: "
+		elog(ERROR, "query-specified return tuple not valid for Connectby: "
 			 "fourth column must be type %s", format_type_be(INT4OID));
 
 	/* OK, the tupdesc is valid for our purposes */
