@@ -8,7 +8,7 @@
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/access/gist/gistscan.c,v 1.65 2006/10/04 00:29:48 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/access/gist/gistscan.c,v 1.65.2.1 2008/08/23 10:41:38 teodor Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -67,6 +67,7 @@ gistrescan(PG_FUNCTION_ARGS)
 			ReleaseBuffer(so->markbuf);
 			so->markbuf = InvalidBuffer;
 		}
+
 	}
 	else
 	{
@@ -82,6 +83,8 @@ gistrescan(PG_FUNCTION_ARGS)
 		scan->opaque = so;
 	}
 
+	so->nPageData = so->curPageData = 0;
+
 	/* Update scan key, if a new one is given */
 	if (key && scan->numberOfKeys > 0)
 	{
@@ -148,6 +151,11 @@ gistmarkpos(PG_FUNCTION_ARGS)
 		so->markbuf = so->curbuf;
 	}
 
+	so->markNPageData = so->nPageData;
+	so->markCurPageData = so->curPageData;
+	if ( so->markNPageData > 0 )
+		memcpy( so->markPageData, so->pageData, sizeof(ItemPointerData) * so->markNPageData );		
+
 	PG_RETURN_VOID();
 }
 
@@ -197,6 +205,11 @@ gistrestrpos(PG_FUNCTION_ARGS)
 		so->curbuf = so->markbuf;
 	}
 
+	so->nPageData = so->markNPageData;
+	so->curPageData = so->markNPageData;
+	if ( so->markNPageData > 0 )
+		memcpy( so->pageData, so->markPageData, sizeof(ItemPointerData) * so->markNPageData );		
+
 	PG_RETURN_VOID();
 }
 
