@@ -3,12 +3,12 @@
  * allpaths.c
  *	  Routines to find possible search paths for processing a query
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/optimizer/path/allpaths.c,v 1.108.2.1 2003/12/17 17:08:06 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/optimizer/path/allpaths.c,v 1.154.2.2 2008/01/11 04:02:25 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -27,8 +27,9 @@
 #include "optimizer/planner.h"
 #include "optimizer/prep.h"
 #include "optimizer/var.h"
-#include "parser/parsetree.h"
 #include "parser/parse_clause.h"
+#include "parser/parse_expr.h"
+#include "parser/parsetree.h"
 #include "rewrite/rewriteManip.h"
 
 
@@ -37,17 +38,20 @@ bool		enable_geqo = false;	/* just in case GUC doesn't set it */
 int			geqo_threshold;
 
 
-static void set_base_rel_pathlists(Query *root);
-static void set_plain_rel_pathlist(Query *root, RelOptInfo *rel,
+static void set_base_rel_pathlists(PlannerInfo *root);
+static void set_rel_pathlist(PlannerInfo *root, RelOptInfo *rel, Index rti);
+static void set_plain_rel_pathlist(PlannerInfo *root, RelOptInfo *rel,
 					   RangeTblEntry *rte);
-static void set_inherited_rel_pathlist(Query *root, RelOptInfo *rel,
-						   Index rti, RangeTblEntry *rte,
-						   List *inheritlist);
-static void set_subquery_pathlist(Query *root, RelOptInfo *rel,
+static void set_append_rel_pathlist(PlannerInfo *root, RelOptInfo *rel,
+						Index rti, RangeTblEntry *rte);
+static void set_subquery_pathlist(PlannerInfo *root, RelOptInfo *rel,
 					  Index rti, RangeTblEntry *rte);
-static void set_function_pathlist(Query *root, RelOptInfo *rel,
+static void set_function_pathlist(PlannerInfo *root, RelOptInfo *rel,
 					  RangeTblEntry *rte);
-static RelOptInfo *make_one_rel_by_joins(Query *root, int levels_needed,
+static void set_values_pathlist(PlannerInfo *root, RelOptInfo *rel,
+					RangeTblEntry *rte);
+static RelOptInfo *make_rel_from_joinlist(PlannerInfo *root, List *joinlist);
+static RelOptInfo *make_one_rel_by_joins(PlannerInfo *root, int levels_needed,
 					  List *initial_rels);
 static bool subquery_is_pushdown_safe(Query *subquery, Query *topquery,
 						  bool *differentTypes);
@@ -57,9 +61,10 @@ static void compare_tlist_datatypes(List *tlist, List *colTypes,
 						bool *differentTypes);
 static bool qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
 					  bool *differentTypes);
-static void subquery_push_qual(Query *subquery, Index rti, Node *qual);
+static void subquery_push_qual(Query *subquery,
+				   RangeTblEntry *rte, Index rti, Node *qual);
 static void recurse_push_qual(Node *setOp, Query *topquery,
-				  Index rti, Node *qual);
+				  RangeTblEntry *rte, Index rti, Node *qual);
 
 
 /*
@@ -68,7 +73,7 @@ static void recurse_push_qual(Node *setOp, Query *topquery,
  *	  single rel that represents the join of all base rels in the query.
  */
 RelOptInfo *
-make_one_rel(Query *root)
+make_one_rel(PlannerInfo *root, List *joinlist)
 {
 	RelOptInfo *rel;
 
@@ -80,14 +85,36 @@ make_one_rel(Query *root)
 	/*
 	 * Generate access paths for the entire join tree.
 	 */
-	Assert(root->jointree != NULL && IsA(root->jointree, FromExpr));
-
-	rel = make_fromexpr_rel(root, root->jointree);
+	rel = make_rel_from_joinlist(root, joinlist);
 
 	/*
-	 * The result should join all the query's base rels.
+	 * The result should join all and only the query's base rels.
 	 */
-	Assert(bms_num_members(rel->relids) == length(root->base_rel_list));
+#ifdef USE_ASSERT_CHECKING
+	{
+		int			num_base_rels = 0;
+		Index		rti;
+
+		for (rti = 1; rti < root->simple_rel_array_size; rti++)
+		{
+			RelOptInfo *brel = root->simple_rel_array[rti];
+
+			if (brel == NULL)
+				continue;
+
+			Assert(brel->relid == rti); /* sanity check on array */
+
+			/* ignore RTEs that are "other rels" */
+			if (brel->reloptkind != RELOPT_BASEREL)
+				continue;
+
+			Assert(bms_is_member(rti, rel->relids));
+			num_base_rels++;
+		}
+
+		Assert(bms_num_members(rel->relids) == num_base_rels);
+	}
+#endif
 
 	return rel;
 }
@@ -99,46 +126,67 @@ make_one_rel(Query *root)
  *	  Each useful path is attached to its relation's 'pathlist' field.
  */
 static void
-set_base_rel_pathlists(Query *root)
+set_base_rel_pathlists(PlannerInfo *root)
 {
-	List	   *rellist;
+	Index		rti;
 
-	foreach(rellist, root->base_rel_list)
+	for (rti = 1; rti < root->simple_rel_array_size; rti++)
 	{
-		RelOptInfo *rel = (RelOptInfo *) lfirst(rellist);
-		Index		rti = rel->relid;
-		RangeTblEntry *rte;
-		List	   *inheritlist;
+		RelOptInfo *rel = root->simple_rel_array[rti];
 
-		Assert(rti > 0);		/* better be base rel */
-		rte = rt_fetch(rti, root->rtable);
+		/* there may be empty slots corresponding to non-baserel RTEs */
+		if (rel == NULL)
+			continue;
 
-		if (rel->rtekind == RTE_SUBQUERY)
-		{
-			/* Subquery --- generate a separate plan for it */
-			set_subquery_pathlist(root, rel, rti, rte);
-		}
-		else if (rel->rtekind == RTE_FUNCTION)
-		{
-			/* RangeFunction --- generate a separate plan for it */
-			set_function_pathlist(root, rel, rte);
-		}
-		else if ((inheritlist = expand_inherited_rtentry(root, rti, true))
-				 != NIL)
-		{
-			/* Relation is root of an inheritance tree, process specially */
-			set_inherited_rel_pathlist(root, rel, rti, rte, inheritlist);
-		}
-		else
-		{
-			/* Plain relation */
-			set_plain_rel_pathlist(root, rel, rte);
-		}
+		Assert(rel->relid == rti);		/* sanity check on array */
+
+		/* ignore RTEs that are "other rels" */
+		if (rel->reloptkind != RELOPT_BASEREL)
+			continue;
+
+		set_rel_pathlist(root, rel, rti);
+	}
+}
+
+/*
+ * set_rel_pathlist
+ *	  Build access paths for a base relation
+ */
+static void
+set_rel_pathlist(PlannerInfo *root, RelOptInfo *rel, Index rti)
+{
+	RangeTblEntry *rte = rt_fetch(rti, root->parse->rtable);
+
+	if (rte->inh)
+	{
+		/* It's an "append relation", process accordingly */
+		set_append_rel_pathlist(root, rel, rti, rte);
+	}
+	else if (rel->rtekind == RTE_SUBQUERY)
+	{
+		/* Subquery --- generate a separate plan for it */
+		set_subquery_pathlist(root, rel, rti, rte);
+	}
+	else if (rel->rtekind == RTE_FUNCTION)
+	{
+		/* RangeFunction --- generate a separate plan for it */
+		set_function_pathlist(root, rel, rte);
+	}
+	else if (rel->rtekind == RTE_VALUES)
+	{
+		/* Values list --- generate a separate plan for it */
+		set_values_pathlist(root, rel, rte);
+	}
+	else
+	{
+		/* Plain relation */
+		Assert(rel->rtekind == RTE_RELATION);
+		set_plain_rel_pathlist(root, rel, rte);
+	}
 
 #ifdef OPTIMIZER_DEBUG
-		debug_print_rel(root, rel);
+	debug_print_rel(root, rel);
 #endif
-	}
 }
 
 /*
@@ -146,11 +194,40 @@ set_base_rel_pathlists(Query *root)
  *	  Build access paths for a plain relation (no subquery, no inheritance)
  */
 static void
-set_plain_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
+set_plain_rel_pathlist(PlannerInfo *root, RelOptInfo *rel, RangeTblEntry *rte)
 {
 	/* Mark rel with estimated output rows, width, etc */
 	set_baserel_size_estimates(root, rel);
 
+	/* Test any partial indexes of rel for applicability */
+	check_partial_indexes(root, rel);
+
+	/*
+	 * Check to see if we can extract any restriction conditions from join
+	 * quals that are OR-of-AND structures.  If so, add them to the rel's
+	 * restriction list, and recompute the size estimates.
+	 */
+	if (create_or_index_quals(root, rel))
+		set_baserel_size_estimates(root, rel);
+
+	/*
+	 * If we can prove we don't need to scan the rel via constraint exclusion,
+	 * set up a single dummy path for it.  (Rather than inventing a special
+	 * "dummy" path type, we represent this as an AppendPath with no members.)
+	 */
+	if (relation_excluded_by_constraints(rel, rte))
+	{
+		/* Reset output-rows estimate to 0 */
+		rel->rows = 0;
+
+		add_path(rel, (Path *) create_append_path(rel, NIL));
+
+		/* Select cheapest path (pretty easy in this case...) */
+		set_cheapest(rel);
+
+		return;
+	}
+
 	/*
 	 * Generate paths and add them to the rel's pathlist.
 	 *
@@ -162,152 +239,160 @@ set_plain_rel_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
 	/* Consider sequential scan */
 	add_path(rel, create_seqscan_path(root, rel));
 
-	/* Consider TID scans */
-	create_tidscan_paths(root, rel);
-
-	/* Consider index paths for both simple and OR index clauses */
+	/* Consider index scans */
 	create_index_paths(root, rel);
 
-	/* create_index_paths must be done before create_or_index_paths */
-	create_or_index_paths(root, rel);
+	/* Consider TID scans */
+	create_tidscan_paths(root, rel);
 
 	/* Now find the cheapest of the paths for this rel */
 	set_cheapest(rel);
 }
 
 /*
- * set_inherited_rel_pathlist
- *	  Build access paths for a inheritance tree rooted at rel
+ * set_append_rel_pathlist
+ *	  Build access paths for an "append relation"
  *
- * inheritlist is a list of RT indexes of all tables in the inheritance tree,
- * including a duplicate of the parent itself.	Note we will not come here
- * unless there's at least one child in addition to the parent.
- *
- * NOTE: the passed-in rel and RTE will henceforth represent the appended
- * result of the whole inheritance tree.  The members of inheritlist represent
- * the individual tables --- in particular, the inheritlist member that is a
- * duplicate of the parent RTE represents the parent table alone.
- * We will generate plans to scan the individual tables that refer to
- * the inheritlist RTEs, whereas Vars elsewhere in the plan tree that
- * refer to the original RTE are taken to refer to the append output.
- * In particular, this means we have separate RelOptInfos for the parent
- * table and for the append output, which is a good thing because they're
- * not the same size.
+ * The passed-in rel and RTE represent the entire append relation.	The
+ * relation's contents are computed by appending together the output of
+ * the individual member relations.  Note that in the inheritance case,
+ * the first member relation is actually the same table as is mentioned in
+ * the parent RTE ... but it has a different RTE and RelOptInfo.  This is
+ * a good thing because their outputs are not the same size.
  */
 static void
-set_inherited_rel_pathlist(Query *root, RelOptInfo *rel,
-						   Index rti, RangeTblEntry *rte,
-						   List *inheritlist)
+set_append_rel_pathlist(PlannerInfo *root, RelOptInfo *rel,
+						Index rti, RangeTblEntry *rte)
 {
 	int			parentRTindex = rti;
-	Oid			parentOID = rte->relid;
 	List	   *subpaths = NIL;
-	List	   *il;
+	ListCell   *l;
 
 	/*
-	 * XXX for now, can't handle inherited expansion of FOR UPDATE; can we
-	 * do better?
+	 * XXX for now, can't handle inherited expansion of FOR UPDATE/SHARE; can
+	 * we do better?  (This will take some redesign because the executor
+	 * currently supposes that every rowMark relation is involved in every row
+	 * returned by the query.)
 	 */
-	if (intMember(parentRTindex, root->rowMarks))
+	if (get_rowmark(root->parse, parentRTindex))
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("SELECT FOR UPDATE is not supported for inheritance queries")));
+				 errmsg("SELECT FOR UPDATE/SHARE is not supported for inheritance queries")));
 
 	/*
-	 * The executor will check the parent table's access permissions when
-	 * it examines the parent's inheritlist entry.  There's no need to
-	 * check twice, so turn off access check bits in the original RTE.
-	 */
-	rte->checkForRead = false;
-	rte->checkForWrite = false;
-
-	/*
-	 * Initialize to compute size estimates for whole inheritance tree
+	 * Initialize to compute size estimates for whole append relation
 	 */
 	rel->rows = 0;
 	rel->width = 0;
 
 	/*
-	 * Generate access paths for each table in the tree (parent AND
-	 * children), and pick the cheapest path for each table.
+	 * Generate access paths for each member relation, and pick the cheapest
+	 * path for each one.
 	 */
-	foreach(il, inheritlist)
+	foreach(l, root->append_rel_list)
 	{
-		int			childRTindex = lfirsti(il);
-		RangeTblEntry *childrte;
-		Oid			childOID;
+		AppendRelInfo *appinfo = (AppendRelInfo *) lfirst(l);
+		int			childRTindex;
 		RelOptInfo *childrel;
-		List	   *reltlist;
-		List	   *parentvars;
-		List	   *childvars;
+		Path	   *childpath;
+		ListCell   *parentvars;
+		ListCell   *childvars;
 
-		childrte = rt_fetch(childRTindex, root->rtable);
-		childOID = childrte->relid;
+		/* append_rel_list contains all append rels; ignore others */
+		if (appinfo->parent_relid != parentRTindex)
+			continue;
+
+		childRTindex = appinfo->child_relid;
 
 		/*
-		 * Make a RelOptInfo for the child so we can do planning.  Do NOT
-		 * attach the RelOptInfo to the query's base_rel_list, however,
-		 * since the child is not part of the main join tree.  Instead,
-		 * the child RelOptInfo is added to other_rel_list.
+		 * The child rel's RelOptInfo was already created during
+		 * add_base_rels_to_query.
 		 */
-		childrel = build_other_rel(root, childRTindex);
+		childrel = find_base_rel(root, childRTindex);
+		Assert(childrel->reloptkind == RELOPT_OTHER_MEMBER_REL);
 
 		/*
-		 * Copy the parent's targetlist and restriction quals to the
-		 * child, with attribute-number adjustment as needed.  We don't
-		 * bother to copy the join quals, since we can't do any joining of
-		 * the individual tables.  Also, we just zap attr_needed rather
-		 * than trying to adjust it; it won't be looked at in the child.
+		 * Copy the parent's targetlist and quals to the child, with
+		 * appropriate substitution of variables.
 		 */
-		reltlist = FastListValue(&rel->reltargetlist);
-		reltlist = (List *)
-			adjust_inherited_attrs((Node *) reltlist,
-								   parentRTindex,
-								   parentOID,
-								   childRTindex,
-								   childOID);
-		FastListFromList(&childrel->reltargetlist, reltlist);
-		childrel->attr_needed = NULL;
+		childrel->reltargetlist = (List *)
+			adjust_appendrel_attrs((Node *) rel->reltargetlist,
+								   appinfo);
 		childrel->baserestrictinfo = (List *)
-			adjust_inherited_attrs((Node *) rel->baserestrictinfo,
-								   parentRTindex,
-								   parentOID,
-								   childRTindex,
-								   childOID);
+			adjust_appendrel_attrs((Node *) rel->baserestrictinfo,
+								   appinfo);
+		childrel->joininfo = (List *)
+			adjust_appendrel_attrs((Node *) rel->joininfo,
+								   appinfo);
 
 		/*
-		 * Now compute child access paths, and save the cheapest.
+		 * Copy the parent's attr_needed data as well, with appropriate
+		 * adjustment of relids and attribute numbers.
 		 */
-		set_plain_rel_pathlist(root, childrel, childrte);
-
-		subpaths = lappend(subpaths, childrel->cheapest_total_path);
+		pfree(childrel->attr_needed);
+		childrel->attr_needed =
+			adjust_appendrel_attr_needed(rel, appinfo,
+										 childrel->min_attr,
+										 childrel->max_attr);
 
 		/*
-		 * Propagate size information from the child back to the parent.
-		 * For simplicity, we use the largest widths from any child as the
-		 * parent estimates.
+		 * Compute the child's access paths, and add the cheapest one to the
+		 * Append path we are constructing for the parent.
+		 *
+		 * It's possible that the child is itself an appendrel, in which case
+		 * we can "cut out the middleman" and just add its child paths to our
+		 * own list.  (We don't try to do this earlier because we need to
+		 * apply both levels of transformation to the quals.) This test also
+		 * handles the case where the child rel need not be scanned because of
+		 * constraint exclusion: it'll have an Append path with no subpaths,
+		 * and will vanish from our list.
+		 */
+		set_rel_pathlist(root, childrel, childRTindex);
+
+		childpath = childrel->cheapest_total_path;
+		if (IsA(childpath, AppendPath))
+			subpaths = list_concat(subpaths,
+								   ((AppendPath *) childpath)->subpaths);
+		else
+			subpaths = lappend(subpaths, childpath);
+
+		/*
+		 * Propagate size information from the child back to the parent. For
+		 * simplicity, we use the largest widths from any child as the parent
+		 * estimates.
 		 */
 		rel->rows += childrel->rows;
 		if (childrel->width > rel->width)
 			rel->width = childrel->width;
 
-		childvars = FastListValue(&childrel->reltargetlist);
-		foreach(parentvars, FastListValue(&rel->reltargetlist))
+		forboth(parentvars, rel->reltargetlist,
+				childvars, childrel->reltargetlist)
 		{
 			Var		   *parentvar = (Var *) lfirst(parentvars);
 			Var		   *childvar = (Var *) lfirst(childvars);
-			int			parentndx = parentvar->varattno - rel->min_attr;
-			int			childndx = childvar->varattno - childrel->min_attr;
 
-			if (childrel->attr_widths[childndx] > rel->attr_widths[parentndx])
-				rel->attr_widths[parentndx] = childrel->attr_widths[childndx];
-			childvars = lnext(childvars);
+			if (IsA(parentvar, Var) &&
+				IsA(childvar, Var))
+			{
+				int			pndx = parentvar->varattno - rel->min_attr;
+				int			cndx = childvar->varattno - childrel->min_attr;
+
+				if (childrel->attr_widths[cndx] > rel->attr_widths[pndx])
+					rel->attr_widths[pndx] = childrel->attr_widths[cndx];
+			}
 		}
 	}
 
 	/*
-	 * Finally, build Append path and install it as the only access path
-	 * for the parent rel.
+	 * Set "raw tuples" count equal to "rows" for the appendrel; needed
+	 * because some places assume rel->tuples is valid for any baserel.
+	 */
+	rel->tuples = rel->rows;
+
+	/*
+	 * Finally, build Append path and install it as the only access path for
+	 * the parent rel.	(Note: this is correct even if we have zero or one
+	 * live subpath due to constraint exclusion.)
 	 */
 	add_path(rel, (Path *) create_append_path(rel, subpaths));
 
@@ -315,36 +400,60 @@ set_inherited_rel_pathlist(Query *root, RelOptInfo *rel,
 	set_cheapest(rel);
 }
 
+/* quick-and-dirty test to see if any joining is needed */
+static bool
+has_multiple_baserels(PlannerInfo *root)
+{
+	int			num_base_rels = 0;
+	Index		rti;
+
+	for (rti = 1; rti < root->simple_rel_array_size; rti++)
+	{
+		RelOptInfo *brel = root->simple_rel_array[rti];
+
+		if (brel == NULL)
+			continue;
+
+		/* ignore RTEs that are "other rels" */
+		if (brel->reloptkind == RELOPT_BASEREL)
+			if (++num_base_rels > 1)
+				return true;
+	}
+	return false;
+}
+
 /*
  * set_subquery_pathlist
  *		Build the (single) access path for a subquery RTE
  */
 static void
-set_subquery_pathlist(Query *root, RelOptInfo *rel,
+set_subquery_pathlist(PlannerInfo *root, RelOptInfo *rel,
 					  Index rti, RangeTblEntry *rte)
 {
+	Query	   *parse = root->parse;
 	Query	   *subquery = rte->subquery;
 	bool	   *differentTypes;
+	double		tuple_fraction;
 	List	   *pathkeys;
+	List	   *subquery_pathkeys;
 
 	/* We need a workspace for keeping track of set-op type coercions */
 	differentTypes = (bool *)
-		palloc0((length(subquery->targetList) + 1) * sizeof(bool));
+		palloc0((list_length(subquery->targetList) + 1) * sizeof(bool));
 
 	/*
 	 * If there are any restriction clauses that have been attached to the
-	 * subquery relation, consider pushing them down to become HAVING
-	 * quals of the subquery itself.  (Not WHERE clauses, since they may
-	 * refer to subquery outputs that are aggregate results.  But
-	 * planner.c will transfer them into the subquery's WHERE if they do
-	 * not.)  This transformation is useful because it may allow us to
-	 * generate a better plan for the subquery than evaluating all the
-	 * subquery output rows and then filtering them.
+	 * subquery relation, consider pushing them down to become WHERE or HAVING
+	 * quals of the subquery itself.  This transformation is useful because it
+	 * may allow us to generate a better plan for the subquery than evaluating
+	 * all the subquery output rows and then filtering them.
 	 *
-	 * There are several cases where we cannot push down clauses.
-	 * Restrictions involving the subquery are checked by
-	 * subquery_is_pushdown_safe().  Restrictions on individual clauses
-	 * are checked by qual_is_pushdown_safe().
+	 * There are several cases where we cannot push down clauses. Restrictions
+	 * involving the subquery are checked by subquery_is_pushdown_safe().
+	 * Restrictions on individual clauses are checked by
+	 * qual_is_pushdown_safe().  Also, we don't want to push down
+	 * pseudoconstant clauses; better to have the gating node above the
+	 * subquery.
 	 *
 	 * Non-pushed-down clauses will get evaluated as qpquals of the
 	 * SubqueryScan node.
@@ -357,17 +466,18 @@ set_subquery_pathlist(Query *root, RelOptInfo *rel,
 	{
 		/* OK to consider pushing down individual quals */
 		List	   *upperrestrictlist = NIL;
-		List	   *lst;
+		ListCell   *l;
 
-		foreach(lst, rel->baserestrictinfo)
+		foreach(l, rel->baserestrictinfo)
 		{
-			RestrictInfo *rinfo = (RestrictInfo *) lfirst(lst);
+			RestrictInfo *rinfo = (RestrictInfo *) lfirst(l);
 			Node	   *clause = (Node *) rinfo->clause;
 
-			if (qual_is_pushdown_safe(subquery, rti, clause, differentTypes))
+			if (!rinfo->pseudoconstant &&
+				qual_is_pushdown_safe(subquery, rti, clause, differentTypes))
 			{
 				/* Push it down */
-				subquery_push_qual(subquery, rti, clause);
+				subquery_push_qual(subquery, rte, rti, clause);
 			}
 			else
 			{
@@ -380,8 +490,25 @@ set_subquery_pathlist(Query *root, RelOptInfo *rel,
 
 	pfree(differentTypes);
 
+	/*
+	 * We can safely pass the outer tuple_fraction down to the subquery if the
+	 * outer level has no joining, aggregation, or sorting to do. Otherwise
+	 * we'd better tell the subquery to plan for full retrieval. (XXX This
+	 * could probably be made more intelligent ...)
+	 */
+	if (parse->hasAggs ||
+		parse->groupClause ||
+		parse->havingQual ||
+		parse->distinctClause ||
+		parse->sortClause ||
+		has_multiple_baserels(root))
+		tuple_fraction = 0.0;	/* default case */
+	else
+		tuple_fraction = root->tuple_fraction;
+
 	/* Generate the plan for the subquery */
-	rel->subplan = subquery_planner(subquery, 0.0 /* default case */ );
+	rel->subplan = subquery_planner(subquery, tuple_fraction,
+									&subquery_pathkeys);
 
 	/* Copy number of output rows from subplan */
 	rel->tuples = rel->subplan->plan_rows;
@@ -390,7 +517,7 @@ set_subquery_pathlist(Query *root, RelOptInfo *rel,
 	set_baserel_size_estimates(root, rel);
 
 	/* Convert subquery pathkeys to outer representation */
-	pathkeys = build_subquery_pathkeys(root, rel, subquery);
+	pathkeys = convert_subquery_pathkeys(root, rel, subquery_pathkeys);
 
 	/* Generate appropriate path */
 	add_path(rel, create_subqueryscan_path(rel, pathkeys));
@@ -404,7 +531,7 @@ set_subquery_pathlist(Query *root, RelOptInfo *rel,
  *		Build the (single) access path for a function RTE
  */
 static void
-set_function_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
+set_function_pathlist(PlannerInfo *root, RelOptInfo *rel, RangeTblEntry *rte)
 {
 	/* Mark rel with estimated output rows, width, etc */
 	set_function_size_estimates(root, rel);
@@ -417,52 +544,96 @@ set_function_pathlist(Query *root, RelOptInfo *rel, RangeTblEntry *rte)
 }
 
 /*
- * make_fromexpr_rel
- *	  Build access paths for a FromExpr jointree node.
+ * set_values_pathlist
+ *		Build the (single) access path for a VALUES RTE
  */
-RelOptInfo *
-make_fromexpr_rel(Query *root, FromExpr *from)
+static void
+set_values_pathlist(PlannerInfo *root, RelOptInfo *rel, RangeTblEntry *rte)
+{
+	/* Mark rel with estimated output rows, width, etc */
+	set_values_size_estimates(root, rel);
+
+	/* Generate appropriate path */
+	add_path(rel, create_valuesscan_path(root, rel));
+
+	/* Select cheapest path (pretty easy in this case...) */
+	set_cheapest(rel);
+}
+
+/*
+ * make_rel_from_joinlist
+ *	  Build access paths using a "joinlist" to guide the join path search.
+ *
+ * See comments for deconstruct_jointree() for definition of the joinlist
+ * data structure.
+ */
+static RelOptInfo *
+make_rel_from_joinlist(PlannerInfo *root, List *joinlist)
 {
 	int			levels_needed;
-	List	   *initial_rels = NIL;
-	List	   *jt;
+	List	   *initial_rels;
+	ListCell   *jl;
 
 	/*
-	 * Count the number of child jointree nodes.  This is the depth of the
-	 * dynamic-programming algorithm we must employ to consider all ways
-	 * of joining the child nodes.
+	 * Count the number of child joinlist nodes.  This is the depth of the
+	 * dynamic-programming algorithm we must employ to consider all ways of
+	 * joining the child nodes.
 	 */
-	levels_needed = length(from->fromlist);
+	levels_needed = list_length(joinlist);
 
 	if (levels_needed <= 0)
 		return NULL;			/* nothing to do? */
 
 	/*
-	 * Construct a list of rels corresponding to the child jointree nodes.
+	 * Construct a list of rels corresponding to the child joinlist nodes.
 	 * This may contain both base rels and rels constructed according to
-	 * explicit JOIN directives.
+	 * sub-joinlists.
 	 */
-	foreach(jt, from->fromlist)
+	initial_rels = NIL;
+	foreach(jl, joinlist)
 	{
-		Node	   *jtnode = (Node *) lfirst(jt);
+		Node	   *jlnode = (Node *) lfirst(jl);
+		RelOptInfo *thisrel;
 
-		initial_rels = lappend(initial_rels,
-							   make_jointree_rel(root, jtnode));
+		if (IsA(jlnode, RangeTblRef))
+		{
+			int			varno = ((RangeTblRef *) jlnode)->rtindex;
+
+			thisrel = find_base_rel(root, varno);
+		}
+		else if (IsA(jlnode, List))
+		{
+			/* Recurse to handle subproblem */
+			thisrel = make_rel_from_joinlist(root, (List *) jlnode);
+		}
+		else
+		{
+			elog(ERROR, "unrecognized joinlist node type: %d",
+				 (int) nodeTag(jlnode));
+			thisrel = NULL;		/* keep compiler quiet */
+		}
+
+		initial_rels = lappend(initial_rels, thisrel);
 	}
 
 	if (levels_needed == 1)
 	{
 		/*
-		 * Single jointree node, so we're done.
+		 * Single joinlist node, so we're done.
 		 */
-		return (RelOptInfo *) lfirst(initial_rels);
+		return (RelOptInfo *) linitial(initial_rels);
 	}
 	else
 	{
 		/*
 		 * Consider the different orders in which we could join the rels,
 		 * using either GEQO or regular optimizer.
+		 *
+		 * We put the initial_rels list into a PlannerInfo field because
+		 * has_legal_joinclause() needs to look at it (ugly :-().
 		 */
+		root->initial_rels = initial_rels;
+
 		if (enable_geqo && levels_needed >= geqo_threshold)
 			return geqo(root, levels_needed, initial_rels);
 		else
@@ -485,18 +656,18 @@ make_fromexpr_rel(Query *root, FromExpr *from)
  * the result of joining all the original relations together.
  */
 static RelOptInfo *
-make_one_rel_by_joins(Query *root, int levels_needed, List *initial_rels)
+make_one_rel_by_joins(PlannerInfo *root, int levels_needed, List *initial_rels)
 {
 	List	  **joinitems;
 	int			lev;
 	RelOptInfo *rel;
 
 	/*
-	 * We employ a simple "dynamic programming" algorithm: we first find
-	 * all ways to build joins of two jointree items, then all ways to
-	 * build joins of three items (from two-item joins and single items),
-	 * then four-item joins, and so on until we have considered all ways
-	 * to join all the items into one rel.
+	 * We employ a simple "dynamic programming" algorithm: we first find all
+	 * ways to build joins of two jointree items, then all ways to build joins
+	 * of three items (from two-item joins and single items), then four-item
+	 * joins, and so on until we have considered all ways to join all the
+	 * items into one rel.
 	 *
 	 * joinitems[j] is a list of all the j-item rels.  Initially we set
 	 * joinitems[1] to represent all the single-jointree-item relations.
@@ -507,7 +678,7 @@ make_one_rel_by_joins(Query *root, int levels_needed, List *initial_rels)
 
 	for (lev = 2; lev <= levels_needed; lev++)
 	{
-		List	   *x;
+		ListCell   *x;
 
 		/*
 		 * Determine all possible pairs of relations to be joined at this
@@ -523,16 +694,6 @@ make_one_rel_by_joins(Query *root, int levels_needed, List *initial_rels)
 		{
 			rel = (RelOptInfo *) lfirst(x);
 
-#ifdef NOT_USED
-
-			/*
-			 * * for each expensive predicate in each path in each
-			 * distinct rel, * consider doing pullup  -- JMH
-			 */
-			if (XfuncMode != XFUNC_NOPULL && XfuncMode != XFUNC_OFF)
-				xfunc_trypullup(rel);
-#endif
-
 			/* Find and save the cheapest paths for this rel */
 			set_cheapest(rel);
 
@@ -547,9 +708,9 @@ make_one_rel_by_joins(Query *root, int levels_needed, List *initial_rels)
 	 */
 	if (joinitems[levels_needed] == NIL)
 		elog(ERROR, "failed to build any %d-way joins", levels_needed);
-	Assert(length(joinitems[levels_needed]) == 1);
+	Assert(list_length(joinitems[levels_needed]) == 1);
 
-	rel = (RelOptInfo *) lfirst(joinitems[levels_needed]);
+	rel = (RelOptInfo *) linitial(joinitems[levels_needed]);
 
 	return rel;
 }
@@ -656,26 +817,31 @@ recurse_pushdown_safe(Node *setOp, Query *topquery,
  * Compare tlist's datatypes against the list of set-operation result types.
  * For any items that are different, mark the appropriate element of
  * differentTypes[] to show that this column will have type conversions.
+ *
+ * We don't have to care about typmods here: the only allowed difference
+ * between set-op input and output typmods is input is a specific typmod
+ * and output is -1, and that does not require a coercion.
  */
 static void
 compare_tlist_datatypes(List *tlist, List *colTypes,
 						bool *differentTypes)
 {
-	List	   *i;
+	ListCell   *l;
+	ListCell   *colType = list_head(colTypes);
 
-	foreach(i, tlist)
+	foreach(l, tlist)
 	{
-		TargetEntry *tle = (TargetEntry *) lfirst(i);
+		TargetEntry *tle = (TargetEntry *) lfirst(l);
 
-		if (tle->resdom->resjunk)
+		if (tle->resjunk)
 			continue;			/* ignore resjunk columns */
-		if (colTypes == NIL)
+		if (colType == NULL)
 			elog(ERROR, "wrong number of tlist entries");
-		if (tle->resdom->restype != lfirsto(colTypes))
-			differentTypes[tle->resdom->resno] = true;
-		colTypes = lnext(colTypes);
+		if (exprType((Node *) tle->expr) != lfirst_oid(colType))
+			differentTypes[tle->resno] = true;
+		colType = lnext(colType);
 	}
-	if (colTypes != NIL)
+	if (colType != NULL)
 		elog(ERROR, "wrong number of tlist entries");
 }
 
@@ -691,11 +857,14 @@ compare_tlist_datatypes(List *tlist, List *colTypes,
  * it will work correctly: sublinks will already have been transformed into
  * subplans in the qual, but not in the subquery).
  *
- * 2. The qual must not refer to any subquery output columns that were
+ * 2. The qual must not refer to the whole-row output of the subquery
+ * (since there is no easy way to name that within the subquery itself).
+ *
+ * 3. The qual must not refer to any subquery output columns that were
  * found to have inconsistent types across a set operation tree by
  * subquery_is_pushdown_safe().
  *
- * 3. If the subquery uses DISTINCT ON, we must not push down any quals that
+ * 4. If the subquery uses DISTINCT ON, we must not push down any quals that
  * refer to non-DISTINCT output columns, because that could change the set
  * of rows returned.  This condition is vacuous for DISTINCT, because then
  * there are no non-DISTINCT output columns, but unfortunately it's fairly
@@ -703,9 +872,13 @@ compare_tlist_datatypes(List *tlist, List *colTypes,
  * parsetree representation.  It's cheaper to just make sure all the Vars
  * in the qual refer to DISTINCT columns.
  *
- * 4. We must not push down any quals that refer to subselect outputs that
+ * 5. We must not push down any quals that refer to subselect outputs that
  * return sets, else we'd introduce functions-returning-sets into the
  * subquery's WHERE/HAVING quals.
+ *
+ * 6. We must not push down any quals that refer to subselect outputs that
+ * contain volatile functions, for fear of introducing strange results due
+ * to multiple evaluation of a volatile function.
  */
 static bool
 qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
@@ -713,7 +886,7 @@ qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
 {
 	bool		safe = true;
 	List	   *vars;
-	List	   *vl;
+	ListCell   *vl;
 	Bitmapset  *tested = NULL;
 
 	/* Refuse subselects (point 1) */
@@ -721,8 +894,8 @@ qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
 		return false;
 
 	/*
-	 * Examine all Vars used in clause; since it's a restriction clause,
-	 * all such Vars must refer to subselect output columns.
+	 * Examine all Vars used in clause; since it's a restriction clause, all
+	 * such Vars must refer to subselect output columns.
 	 */
 	vars = pull_var_clause(qual, false);
 	foreach(vl, vars)
@@ -732,16 +905,23 @@ qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
 
 		Assert(var->varno == rti);
 
+		/* Check point 2 */
+		if (var->varattno == 0)
+		{
+			safe = false;
+			break;
+		}
+
 		/*
-		 * We use a bitmapset to avoid testing the same attno more than
-		 * once.  (NB: this only works because subquery outputs can't have
-		 * negative attnos.)
+		 * We use a bitmapset to avoid testing the same attno more than once.
+		 * (NB: this only works because subquery outputs can't have negative
+		 * attnos.)
 		 */
 		if (bms_is_member(var->varattno, tested))
 			continue;
 		tested = bms_add_member(tested, var->varattno);
 
-		/* Check point 2 */
+		/* Check point 3 */
 		if (differentTypes[var->varattno])
 		{
 			safe = false;
@@ -751,9 +931,9 @@ qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
 		/* Must find the tlist element referenced by the Var */
 		tle = get_tle_by_resno(subquery->targetList, var->varattno);
 		Assert(tle != NULL);
-		Assert(!tle->resdom->resjunk);
+		Assert(!tle->resjunk);
 
-		/* If subquery uses DISTINCT or DISTINCT ON, check point 3 */
+		/* If subquery uses DISTINCT or DISTINCT ON, check point 4 */
 		if (subquery->distinctClause != NIL &&
 			!targetIsInSortList(tle, subquery->distinctClause))
 		{
@@ -762,15 +942,22 @@ qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
 			break;
 		}
 
-		/* Refuse functions returning sets (point 4) */
+		/* Refuse functions returning sets (point 5) */
 		if (expression_returns_set((Node *) tle->expr))
 		{
 			safe = false;
 			break;
 		}
+
+		/* Refuse volatile functions (point 6) */
+		if (contain_volatile_functions((Node *) tle->expr))
+		{
+			safe = false;
+			break;
+		}
 	}
 
-	freeList(vars);
+	list_free(vars);
 	bms_free(tested);
 
 	return safe;
@@ -780,35 +967,44 @@ qual_is_pushdown_safe(Query *subquery, Index rti, Node *qual,
  * subquery_push_qual - push down a qual that we have determined is safe
  */
 static void
-subquery_push_qual(Query *subquery, Index rti, Node *qual)
+subquery_push_qual(Query *subquery, RangeTblEntry *rte, Index rti, Node *qual)
 {
 	if (subquery->setOperations != NULL)
 	{
 		/* Recurse to push it separately to each component query */
-		recurse_push_qual(subquery->setOperations, subquery, rti, qual);
+		recurse_push_qual(subquery->setOperations, subquery,
+						  rte, rti, qual);
 	}
 	else
 	{
 		/*
-		 * We need to replace Vars in the qual (which must refer to
-		 * outputs of the subquery) with copies of the subquery's
-		 * targetlist expressions.	Note that at this point, any uplevel
-		 * Vars in the qual should have been replaced with Params, so they
-		 * need no work.
+		 * We need to replace Vars in the qual (which must refer to outputs of
+		 * the subquery) with copies of the subquery's targetlist expressions.
+		 * Note that at this point, any uplevel Vars in the qual should have
+		 * been replaced with Params, so they need no work.
 		 *
 		 * This step also ensures that when we are pushing into a setop tree,
 		 * each component query gets its own copy of the qual.
 		 */
-		qual = ResolveNew(qual, rti, 0,
+		qual = ResolveNew(qual, rti, 0, rte,
 						  subquery->targetList,
 						  CMD_SELECT, 0);
-		subquery->havingQual = make_and_qual(subquery->havingQual,
-											 qual);
+
+		/*
+		 * Now attach the qual to the proper place: normally WHERE, but if the
+		 * subquery uses grouping or aggregation, put it in HAVING (since the
+		 * qual really refers to the group-result rows).
+		 */
+		if (subquery->hasAggs || subquery->groupClause || subquery->havingQual)
+			subquery->havingQual = make_and_qual(subquery->havingQual, qual);
+		else
+			subquery->jointree->quals =
+				make_and_qual(subquery->jointree->quals, qual);
 
 		/*
 		 * We need not change the subquery's hasAggs or hasSublinks flags,
-		 * since we can't be pushing down any aggregates that weren't
-		 * there before, and we don't push down subselects at all.
+		 * since we can't be pushing down any aggregates that weren't there
+		 * before, and we don't push down subselects at all.
 		 */
 	}
 }
@@ -818,23 +1014,23 @@ subquery_push_qual(Query *subquery, Index rti, Node *qual)
  */
 static void
 recurse_push_qual(Node *setOp, Query *topquery,
-				  Index rti, Node *qual)
+				  RangeTblEntry *rte, Index rti, Node *qual)
 {
 	if (IsA(setOp, RangeTblRef))
 	{
 		RangeTblRef *rtr = (RangeTblRef *) setOp;
-		RangeTblEntry *rte = rt_fetch(rtr->rtindex, topquery->rtable);
-		Query	   *subquery = rte->subquery;
+		RangeTblEntry *subrte = rt_fetch(rtr->rtindex, topquery->rtable);
+		Query	   *subquery = subrte->subquery;
 
 		Assert(subquery != NULL);
-		subquery_push_qual(subquery, rti, qual);
+		subquery_push_qual(subquery, rte, rti, qual);
 	}
 	else if (IsA(setOp, SetOperationStmt))
 	{
 		SetOperationStmt *op = (SetOperationStmt *) setOp;
 
-		recurse_push_qual(op->larg, topquery, rti, qual);
-		recurse_push_qual(op->rarg, topquery, rti, qual);
+		recurse_push_qual(op->larg, topquery, rte, rti, qual);
+		recurse_push_qual(op->rarg, topquery, rte, rti, qual);
 	}
 	else
 	{
@@ -868,22 +1064,22 @@ print_relids(Relids relids)
 }
 
 static void
-print_restrictclauses(Query *root, List *clauses)
+print_restrictclauses(PlannerInfo *root, List *clauses)
 {
-	List	   *l;
+	ListCell   *l;
 
 	foreach(l, clauses)
 	{
 		RestrictInfo *c = lfirst(l);
 
-		print_expr((Node *) c->clause, root->rtable);
+		print_expr((Node *) c->clause, root->parse->rtable);
 		if (lnext(l))
 			printf(", ");
 	}
 }
 
 static void
-print_path(Query *root, Path *path, int indent)
+print_path(PlannerInfo *root, Path *path, int indent)
 {
 	const char *ptype;
 	bool		join = false;
@@ -898,6 +1094,15 @@ print_path(Query *root, Path *path, int indent)
 		case T_IndexPath:
 			ptype = "IdxScan";
 			break;
+		case T_BitmapHeapPath:
+			ptype = "BitmapHeapScan";
+			break;
+		case T_BitmapAndPath:
+			ptype = "BitmapAndPath";
+			break;
+		case T_BitmapOrPath:
+			ptype = "BitmapOrPath";
+			break;
 		case T_TidPath:
 			ptype = "TidScan";
 			break;
@@ -906,7 +1111,6 @@ print_path(Query *root, Path *path, int indent)
 			break;
 		case T_ResultPath:
 			ptype = "Result";
-			subpath = ((ResultPath *) path)->subpath;
 			break;
 		case T_MaterialPath:
 			ptype = "Material";
@@ -950,7 +1154,7 @@ print_path(Query *root, Path *path, int indent)
 		for (i = 0; i < indent; i++)
 			printf("\t");
 		printf("  pathkeys: ");
-		print_pathkeys(path->pathkeys, root->rtable);
+		print_pathkeys(path->pathkeys, root->parse->rtable);
 	}
 
 	if (join)
@@ -986,9 +1190,9 @@ print_path(Query *root, Path *path, int indent)
 }
 
 void
-debug_print_rel(Query *root, RelOptInfo *rel)
+debug_print_rel(PlannerInfo *root, RelOptInfo *rel)
 {
-	List	   *l;
+	ListCell   *l;
 
 	printf("RELOPTINFO (");
 	print_relids(rel->relids);
@@ -1001,14 +1205,10 @@ debug_print_rel(Query *root, RelOptInfo *rel)
 		printf("\n");
 	}
 
-	foreach(l, rel->joininfo)
+	if (rel->joininfo)
 	{
-		JoinInfo   *j = (JoinInfo *) lfirst(l);
-
-		printf("\tjoininfo (");
-		print_relids(j->unjoined_relids);
-		printf("): ");
-		print_restrictclauses(root, j->jinfo_restrictinfo);
+		printf("\tjoininfo: ");
+		print_restrictclauses(root, rel->joininfo);
 		printf("\n");
 	}
 
