@@ -42,7 +42,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/utils/error/elog.c,v 1.178 2006/11/28 12:54:42 petere Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/error/elog.c,v 1.178.2.5 2008/07/08 22:17:55 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -76,6 +76,8 @@ ErrorContextCallback *error_context_stack = NULL;
 
 sigjmp_buf *PG_exception_stack = NULL;
 
+extern bool redirection_done;
+
 /* GUC parameters */
 PGErrorVerbosity Log_error_verbosity = PGERROR_VERBOSE;
 char	   *Log_line_prefix = NULL;		/* format for extra log line info */
@@ -121,7 +123,7 @@ static char *expand_fmt_string(const char *fmt, ErrorData *edata);
 static const char *useful_strerror(int errnum);
 static const char *error_severity(int elevel);
 static void append_with_tabs(StringInfo buf, const char *str);
-
+static void write_pipe_chunks(int fd, char *data, int len);
 
 /*
  * errstart --- begin an error-reporting cycle
@@ -259,10 +261,15 @@ errstart(int elevel, const char *filename, int lineno,
 
 		/*
 		 * If we recurse more than once, the problem might be something broken
-		 * in a context traceback routine.	Abandon them too.
+		 * in a context traceback routine.  Abandon them too.  We also
+		 * abandon attempting to print the error statement (which, if long,
+		 * could itself be the source of the recursive failure).
 		 */
 		if (recursion_depth > 2)
+		{
 			error_context_stack = NULL;
+			debug_query_string = NULL;
+		}
 	}
 	if (++errordata_stack_depth >= ERRORDATA_STACK_SIZE)
 	{
@@ -1198,6 +1205,7 @@ write_syslog(int level, const char *line)
 	static unsigned long seq = 0;
 
 	int			len;
+	const char *nlpos;
 
 	/* Open syslog connection if not done yet */
 	if (!openlog_done)
@@ -1220,17 +1228,17 @@ write_syslog(int level, const char *line)
 	 * fact, it does work around by splitting up messages into smaller pieces.
 	 *
 	 * We divide into multiple syslog() calls if message is too long or if the
-	 * message contains embedded NewLine(s) '\n'.
+	 * message contains embedded newline(s).
 	 */
 	len = strlen(line);
-	if (len > PG_SYSLOG_LIMIT || strchr(line, '\n') != NULL)
+	nlpos = strchr(line, '\n');
+	if (len > PG_SYSLOG_LIMIT || nlpos != NULL)
 	{
 		int			chunk_nr = 0;
 
 		while (len > 0)
 		{
 			char		buf[PG_SYSLOG_LIMIT + 1];
-			const char *nlpos;
 			int			buflen;
 			int			i;
 
@@ -1239,11 +1247,12 @@ write_syslog(int level, const char *line)
 			{
 				line++;
 				len--;
+				/* we need to recompute the next newline's position, too */
+				nlpos = strchr(line, '\n');
 				continue;
 			}
 
 			/* copy one line, or as much as will fit, to buf */
-			nlpos = strchr(line, '\n');
 			if (nlpos != NULL)
 				buflen = nlpos - line;
 			else
@@ -1715,13 +1724,21 @@ send_message_to_server_log(ErrorData *edata)
 		 * anything going there and write it to the eventlog instead.
 		 *
 		 * If stderr redirection is active, it's ok to write to stderr because
-		 * that's really a pipe to the syslogger process.
+		 * that's really a pipe to the syslogger process. Unless we're in the
+		 * postmaster, and the syslogger process isn't started yet.
 		 */
-		if ((!Redirect_stderr || am_syslogger) && pgwin32_is_service())
+		if (pgwin32_is_service() && (!redirection_done || am_syslogger) )
 			write_eventlog(edata->elevel, buf.data);
 		else
 #endif
-			fprintf(stderr, "%s", buf.data);
+			/* only use the chunking protocol if we know the syslogger should
+			 * be catching stderr output, and we are not ourselves the
+			 * syslogger. Otherwise, go directly to stderr.
+			 */
+			if (redirection_done && !am_syslogger)
+				write_pipe_chunks(fileno(stderr), buf.data, buf.len);
+			else
+				write(fileno(stderr), buf.data, buf.len);
 	}
 
 	/* If in the syslogger process, try to write messages direct to file */
@@ -1731,6 +1748,37 @@ send_message_to_server_log(ErrorData *edata)
 	pfree(buf.data);
 }
 
+/*
+ * Send data to the syslogger using the chunked protocol
+ */
+static void
+write_pipe_chunks(int fd, char *data, int len)
+{
+	PipeProtoChunk p;
+
+	Assert(len > 0);
+
+	p.proto.nuls[0] = p.proto.nuls[1] = '\0';
+	p.proto.pid = MyProcPid;
+
+	/* write all but the last chunk */
+	while (len > PIPE_MAX_PAYLOAD)
+	{
+		p.proto.is_last = 'f';
+		p.proto.len = PIPE_MAX_PAYLOAD;
+		memcpy(p.proto.data, data, PIPE_MAX_PAYLOAD);
+		write(fd, &p, PIPE_HEADER_SIZE + PIPE_MAX_PAYLOAD);
+		data += PIPE_MAX_PAYLOAD;
+		len -= PIPE_MAX_PAYLOAD;
+	}
+
+	/* write the last chunk */
+	p.proto.is_last = 't';
+	p.proto.len = len;
+	memcpy(p.proto.data, data, len);
+	write(fd, &p, PIPE_HEADER_SIZE + len);
+}
+
 
 /*
  * Write error report to client
@@ -2053,6 +2101,7 @@ write_stderr(const char *fmt,...)
 #ifndef WIN32
 	/* On Unix, we just fprintf to stderr */
 	vfprintf(stderr, fmt, ap);
+	fflush(stderr);
 #else
 
 	/*
@@ -2068,8 +2117,11 @@ write_stderr(const char *fmt,...)
 		write_eventlog(EVENTLOG_ERROR_TYPE, errbuf);
 	}
 	else
+	{
 		/* Not running as service, write to stderr */
 		vfprintf(stderr, fmt, ap);
+		fflush(stderr);
+	}
 #endif
 	va_end(ap);
 }
